!
!   This is a sample Fortran 90 program that shows how an application can read a
!   congruent special value dataset. This program will read one dataset,
!   "SnowDepth", that contains actual data. Value of some dataset elements
!   should be replaced by special value.
!
!   Then, this program will read a congruent dataset, "SpecialValues", that
!   contains special values. For simplicity, we will assume that both datasets
!   are 5-by-7 array, and their datatypes are float and int,
!   respectively. Also, we assume that there are two special values; 1 and 2.
!   1 represents "WaterFilled" and 2 represents "IceCovered". 
!
!   After reading both datasets, this program will print some human-readable
!   output. If none of special values are set for an element, the actual value
!   stored in the "SnowDepth" will be printed. If some special values are set, a
!   special character will be printed instead. In this example, '/' is for the
!   "WaterFilled" special value, '\' is for the "IceCovered" special value, and
!   'X' is for the overlapped special value.
!
!   For more information, please refer to the following sections from RFC:
!   Special Values in HDF5 http://www.hdfgroup.org/pubs/rfcs/RFC_Special_Values_in_HDF5.pdf
!      Section 4.3.3 "Parallel Special Value Dataset"
!      Section 4.2.2 "Non-Mutually Exclusive Special Values"
!
PROGRAM congruent_read
USE HDF5
IMPLICIT NONE

CHARACTER(LEN=16) :: filename = "congruent.f90.h5" ! File name
CHARACTER(LEN=9) :: dsetname = "SnowDepth" ! Dataset name
CHARACTER(LEN=13) :: svalname = "SpecialValues" ! Congruent dataset name

INTEGER(HID_T) :: fileid
INTEGER(HID_T) :: spaceid
INTEGER(HID_T) :: dsetid

INTEGER(HSIZE_T), DIMENSION(2) :: dimsize = (/7,5/)
REAL, DIMENSION(7,5) :: data
INTEGER, DIMENSION(7,5) :: masks

INTEGER :: i, j, mask, error
CHARACTER :: ch

! initialize the Fortran interface
CALL h5open_f(error)

! open the HDF5 file using read-only mode
CALL h5fopen_f(filename, H5F_ACC_RDONLY_F, fileid, error)

! create a dataspace for both the actual dataset and the congruent dataset
CALL h5screate_simple_f(2, dimsize, spaceid, error)

! read a dataset
! open the HDF5 dataset
CALL h5dopen_f(fileid, dsetname, dsetid, error)
! read data from the HDF5 dataset, and store them in a variable called "data"
CALL h5dread_f(dsetid, H5T_NATIVE_REAL, data, dimsize, error, spaceid, spaceid)
! close the dataset
CALL h5dclose_f(dsetid, error)

! read a congruent dataset
! open the congruent dataset
CALL h5dopen_f(fileid, svalname, dsetid, error)
! read special values and store them in a variable called "masks"
CALL h5dread_f(dsetid, H5T_NATIVE_INTEGER, masks, dimsize, error, spaceid, spaceid)
! close the dataset
CALL h5dclose_f(dsetid, error)

! close the dataspace
CALL h5sclose_f(spaceid, error)

! close the file
CALL h5fclose_f(fileid, error)

! combine the actual dataset and the congruent dataset to get a final result
! this process could be done when the dataset is written, too
! for each row
do i = 1, dimsize(2)
! for each column
do j = 1, dimsize(1)
! for each element
! "mask" is the special value that corresponds to this data
   mask = masks(j,i)
! if special value is not set for this element, just show the actual element
   if ( mask == 0 ) then
      write (*,"(a,f4.1,a)",advance="no") " ", data(j,i), " "
! if special value is marked for this element, show a special character
   else
! if both special value '2' and special value '1' are marked, show 'X'
      if ( IAND(mask, 1) == 1 .and. IAND(mask, 2) == 2 ) then
         ch = 'X'
! if only special value '2' is marked, show '\'
      else if ( IAND(mask, 2) == 2 ) then
         ch = '\\'
! if only special value '1' is marked, show '/'
      else if ( IAND(mask, 1) == 1 ) then
         ch = '/'
      endif
! if you want to merge the special value into the actual data, "data",
! you may want to do something like, data(j,i) = -1
      write (*,"(a,a,a)",advance="no") "  ", ch, "   "
   endif
enddo
write (*,*)
enddo

! close the Fortran interface
CALL h5close_f(error)

END PROGRAM congruent_read

! vi:set ts=4 sw=4 sts=4:
