;
;   This is a sample IDL program that shows how an application can read a
;   congruent special value dataset. This program will read one dataset,
;   "SnowDepth", that contains actual data. Value of some dataset elements
;   should be replaced by special value.
;
;   Then, this program will read a congruent dataset, "SpecialValues", that
;   contains special values. For simplicity, we will assume that both datasets
;   are 5-by-7 array, and their datatypes are float and byte,
;   respectively. Also, we assume that there are two special values; 1 and 2.
;   1 represents "WaterFilled" and 2 represents "IceCovered". 
;
;   After reading both datasets, this program will print some human-readable
;   output. If none of special values are set for an element, the actual value
;   stored in the "SnowDepth" will be printed. If some special values are set, a
;   special character will be printed instead. In this example, '/' is for the
;   "WaterFilled" special value, '\' is for the "IceCovered" special value, and
;   'X' is for the overlapped special value.
;
;   For more information, please refer to the following sections from RFC:
;   Special Values in HDF5 http://www.hdfgroup.org/pubs/rfcs/RFC_Special_Values_in_HDF5.pdf
;      Section 4.3.3 "Parallel Special Value Dataset"
;      Section 4.2.2 "Non-Mutually Exclusive Special Values"
;
filename = 'congruent.idl.h5' ; File name
dsetname = 'SnowDepth' ; Dataset name
svalname = 'SpecialValues' ; Parallel dataset name

dimsize = [5, 7]

; open the HDF5 file
fileid = H5F_OPEN(filename)

; create a dataspace for both the actual dataset and the congruent dataset
spaceid = H5S_CREATE_SIMPLE(dimsize)

; read a dataset
; open the HDF5 dataset
dsetid = H5D_OPEN(fileid, dsetname)
; read data from the HDF5 dataset, and store them in a variable called "data"
data = H5D_READ(dsetid)
; close the dataset
H5D_CLOSE, dsetid

; read a congruent dataset
; open the congruent dataset
dsetid = H5D_OPEN(fileid, svalname)
; read special values and store them in a variable called "masks"
masks = H5D_READ(dsetid)
; close the dataset
H5D_CLOSE, dsetid

; close the dataspace
H5S_CLOSE, spaceid

; close the file
H5F_CLOSE, fileid

; combine the actual dataset and the congruent dataset to get a final result
; this process could be done when the dataset is written, too
; for each row
FOR i=0,dimsize[0]-1 DO BEGIN
; for each column
  FOR j=0,dimsize[1]-1 DO BEGIN
; "mask" is the special value that corresponds to this data
    mask = masks[i,j]
; if special value is not set for this element, just show the actual element
    IF ( mask EQ 0 ) THEN BEGIN
; if special value is marked for this element, show a special character
      PRINT, FORMAT='(%" %4.1f ",$)', data[i,j]
    ENDIF ELSE BEGIN
      ch = ' '
; if both special value '2' and special value '1' are marked, show 'X'
      IF ( (mask AND 3) EQ 3 ) THEN BEGIN
        ch = 'X'
      ENDIF ELSE BEGIN
; if only special value '2' is marked, show '\'
        IF ( (mask AND 2) EQ 2 ) THEN BEGIN
          ch = '\'
        ENDIF ELSE BEGIN
; if only special value '1' is marked. show '/'
          IF ( (mask AND 1) EQ 1 ) THEN BEGIN
            ch = '/'
          ENDIF
        ENDELSE
      ENDELSE
      PRINT, FORMAT='(%"  %s   ",$)', ch
    ENDELSE
  ENDFOR
  PRINT
ENDFOR

END

; vi:set ts=4 sw=4 sts=4:
