/*
   This is a sample C program that shows how an application can write a
   congruent special value dataset. This program will write one dataset,
   "SnowDepth", that contains actual data.

   Then, this program will create another dataset, "SpecialValues". This
   dataset only contains data for special values. This is called a "congruent"
   dataset in the sense that this dataset will have the same rank and dimension
   as the "SnowDepth" dataset. Note that the datatype does not need to be the
   same as the "SnowDepth" dataset. In fact, the datatype of "SnowDepth" is
   float while the datatype of "SpecialValues" is unsigned char.

   For simplicity, we will use a small 5-by-7 array, and "SnowDepth" will be
   filled by meaningless numbers. We will use two special values. Although it
   is not recorded in the file, we assume that the first special value is for
   "WaterFilled" and its value is 1, and the second special value is for
   "IceCovered" and its value is 2. We will not assume that two special values
   are mutually exclusive; in other words, both special values can be set for
   an element. In this example, an optional attribute that describes the meaning of special
   value is added.

   For more information, please refer to the following sections from RFC:
   Special Values in HDF5 http://www.hdfgroup.org/pubs/rfcs/RFC_Special_Values_in_HDF5.pdf
      Section 4.3.3 "Parallel Special Value Dataset"
      Section 4.2.2 "Non-Mutually Exclusive Special Values"
 */

#define H5_USE_16_API
#include "hdf5.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define errorexit()  _errorexit(__LINE__)

static void _errorexit(int line)
{
	fprintf(stderr, "errorexit at %d\n", line);
	exit(1);
}

static void create_file(const char *fname)
{
	hid_t fileid, spaceid, dsetid, attrspace, attrtype, attrid;
	const char *attrdesc;
	hsize_t dimsize[2] = { 5, 7 };
	int i, j, k;

	/* create an HDF5 file */
	if ((fileid = H5Fcreate(fname, H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT)) < 0) errorexit();
	/* create a dataspace for a dataset and a congruent dataset */
	if ((spaceid = H5Screate_simple(2, dimsize, NULL)) < 0) errorexit();

	/* define a dataset */
	{
		float data[5][7];

		/* create an HDF5 dataset that represents actual data */
		if ((dsetid = H5Dcreate(fileid, "SnowDepth", H5T_NATIVE_FLOAT, spaceid, H5P_DEFAULT)) < 0) errorexit();

		/* fill actual data (some of them are garbages)
		        0  1  2  3  4  5  6
		     +----------------------
		   0 |  1  2  3  4  5  6  7
		   1 |  8  9 10 11 12 13 14
		   2 | 15 16 17 18 19 20 21
		   3 | 22 23 24 25 26 27 28
		   4 | 29 30 31 32 33 34 35
		*/
		for (k = i = 0; i < 5; ++i) {
			for (j = 0; j < 7; ++j)
				data[i][j] = ++k;
		}

		/* write actual data to the dataset we just created */
		if (H5Dwrite(dsetid, H5T_NATIVE_FLOAT, H5S_ALL, H5S_ALL, H5P_DEFAULT, data) < 0) errorexit();
		/* close the dataset */
		if (H5Dclose(dsetid) < 0) errorexit();
	}

	/* define a congruent dataset */
	{
		unsigned char data[5][7];

		/* create an HDF5 dataset that represents special values for the dataset created above */
		if ((dsetid = H5Dcreate(fileid, "SpecialValues", H5T_NATIVE_UCHAR, spaceid, H5P_DEFAULT)) < 0) errorexit();

		/* fill non-mutual special values (0: no special values, 1: WaterFilled, 2: IceCovered, 3: both)
		        0  1  2  3  4  5  6
		     +----------------------
		   0 |  3  0  0  1  0  2  1
		   1 |  3  0  0  1  0  2  1
		   2 |  3  0  0  1  0  2  1
		   3 |  3  0  0  1  0  2  1
		   4 |  3  0  0  1  0  2  1
		*/
		for (i = 0; i < 5; ++i) {
			for (j = 0; j < 7; ++j) {
				unsigned char masks = 0;
				/* for each row, special value '1' is marked once every three elements */
				if (j % 3 == 0) masks |= 1;
				/* for each row, special value '2' is marked once every five elements */
				if (j % 5 == 0) masks |= 2;
				data[i][j] = masks;
			}
		}

		/* write special values to the congruent dataset */
		if (H5Dwrite(dsetid, H5T_NATIVE_UCHAR, H5S_ALL, H5S_ALL, H5P_DEFAULT, data) < 0) errorexit();

		/* (optional) write an attribute that describes this dataset, this attribute just helps the user know what the special value represents for. */
		attrdesc = "1: water-filled, 2: ice-covered";
		if ((attrspace = H5Screate(H5S_SCALAR)) < 0) errorexit();
		if ((attrtype = H5Tcopy(H5T_C_S1)) < 0) errorexit();
		if (H5Tset_size(attrtype, strlen(attrdesc)) < 0) errorexit();

		if ((attrid = H5Acreate(dsetid, "Description", attrtype, attrspace, H5P_DEFAULT)) < 0) errorexit();
		if (H5Awrite(attrid, attrtype, attrdesc) < 0) errorexit();
		if (H5Sclose(attrspace) < 0) errorexit();
		if (H5Aclose(attrid) < 0) errorexit();

		/* close the dataset */
		if (H5Dclose(dsetid) < 0) errorexit();
	}

	/* close the dataspace */
	if (H5Sclose(spaceid) < 0) errorexit();
	/* close the file */
	if (H5Fclose(fileid) < 0) errorexit();
}

int main(int argc, char **argv)
{
	create_file("congruent.c.h5");
	return 0;
}

/* vi:set ts=4 sw=4 sts=4: */
