!
!   This is a sample Fortran 90 program that shows how an application can write a
!   congruent special value dataset. This program will write one dataset,
!   "SnowDepth", that contains actual data.
!
!   Then, this program will create another dataset, "SpecialValues". This
!   dataset only contains data for special values. This is called a "congruent"
!   dataset in the sense that this dataset will have the same rank and dimension
!   as the "SnowDepth" dataset. Note that the datatype does not need to be the
!   same as the "SnowDepth" dataset. In fact, the datatype of "SnowDepth" is
!   float while the datatype of "SpecialValues" is int.
!
!   For simplicity, we will use a small 5-by-7 array, and "SnowDepth" will be
!   filled by meaningless numbers. We will use two special values. Although it
!   is not recorded in the file, we assume that the first special value is for
!   "WaterFilled" and its value is 1, and the second special value is for
!   "IceCovered" and its value is 2. We will not assume that two special values
!   are mutually exclusive; in other words, both special values can be set for
!   an element. In this example, an optional attribute that describes the meaning of special
!   value is added.
!
!   For more information, please refer to the following sections from RFC:
!   Special Values in HDF5 http://www.hdfgroup.org/pubs/rfcs/RFC_Special_Values_in_HDF5.pdf
!      Section 4.3.3 "Parallel Special Value Dataset"
!      Section 4.2.2 "Non-Mutually Exclusive Special Values"
!
PROGRAM congruent_write
USE HDF5
IMPLICIT NONE

CHARACTER(LEN=16) :: filename = "congruent.f90.h5" ! File name
CHARACTER(LEN=9) :: dsetname = "SnowDepth" ! Dataset name
CHARACTER(LEN=13) :: svalname = "SpecialValues" ! Congruent dataset name
CHARACTER(LEN=11) :: attrname = "Description" ! Attribute name
CHARACTER(LEN=31) :: attrdesc = "1: water-filled, 2: ice-covered" ! Description of the congruent dataset

INTEGER(HID_T) :: fileid
INTEGER(HID_T) :: spaceid
INTEGER(HID_T) :: dsetid
INTEGER(HID_T) :: attrspace
INTEGER(HID_T) :: attrtype
INTEGER(HID_T) :: attrid

INTEGER(HSIZE_T), DIMENSION(2) :: dimsize = (/7,5/)
REAL, DIMENSION(7,5) :: data
INTEGER, DIMENSION(7,5) :: svaldata
INTEGER(HSIZE_T), DIMENSION(1) :: attrdim

INTEGER :: i, j, k, masks, error

! initialize the Fortran interface
CALL h5open_f(error)

! create an HDF5 file
CALL h5fcreate_f(filename, H5F_ACC_TRUNC_F, fileid, error)

! create a dataspace for a dataset and a congruent dataset
CALL h5screate_simple_f(2, dimsize, spaceid, error)

! define a dataset
! create an HDF5 dataset that represents actual data
CALL h5dcreate_f(fileid, dsetname, H5T_NATIVE_REAL, spaceid, dsetid, error)
! fill actual data
!		        0  1  2  3  4  5  6
!		     +----------------------
!		   0 |  1  2  3  4  5  6  7
!		   1 |  8  9 10 11 12 13 14
!		   2 | 15 16 17 18 19 20 21
!		   3 | 22 23 24 25 26 27 28
!		   4 | 29 30 31 32 33 34 35
k = 1
do i = 1, dimsize(2)
do j = 1, dimsize(1)
   data(j,i) = k
   k = k + 1
enddo
enddo
! write actual data to the dataset we just created
CALL h5dwrite_f(dsetid, H5T_NATIVE_REAL, data, dimsize, error)
! close the dataset
CALL h5dclose_f(dsetid, error)

! define a congruent dataset
! create an HDF5 dataset that represents special values for the dataset created above
CALL h5dcreate_f(fileid, svalname, H5T_NATIVE_INTEGER, spaceid, dsetid, error)
! fill non-mutual special values (0: no special values, 1: WaterFilled, 2: IceCovered, 3: both)
!		        0  1  2  3  4  5  6
!		     +----------------------
!		   0 |  3  0  0  1  0  2  1
!		   1 |  3  0  0  1  0  2  1
!		   2 |  3  0  0  1  0  2  1
!		   3 |  3  0  0  1  0  2  1
!		   4 |  3  0  0  1  0  2  1
do i = 1, dimsize(2)
do j = 1, dimsize(1)
   masks = 0
! for each row, special value '1' is marked once every three elements
   if ( MOD(j - 1, 3) == 0 ) then
      masks = IOR(masks, 1)
   endif
! for each row, special value '2' is marked once every five elements
   if ( MOD(j - 1, 5) == 0 ) then
      masks = IOR(masks, 2)
   endif
   svaldata(j,i) = masks
enddo
enddo
! write special values to the congruent dataset
CALL h5dwrite_f(dsetid, H5T_NATIVE_INTEGER, svaldata, dimsize, error)

! (optional) write an attribute that describes the meaning of the special value
CALL h5screate_f(H5S_SCALAR_F, attrspace, error)
CALL h5tcopy_f(H5T_NATIVE_CHARACTER, attrtype, error)
CALL h5tset_size_f(attrtype, 31, error)
CALL h5acreate_f(dsetid, attrname, attrtype, attrspace, attrid, error)
attrdim(1) = 1
CALL h5awrite_f(attrid, attrtype, attrdesc, attrdim, error)
CALL h5sclose_f(attrspace, error)
CALL h5aclose_f(attrid, error)

! close the dataset
CALL h5dclose_f(dsetid, error)

! close the dataspace
CALL h5sclose_f(spaceid, error)

! close the file
CALL h5fclose_f(fileid, error)

! close the Fortran interface
CALL h5close_f(error)

END PROGRAM congruent_write

! vi:set ts=4 sw=4 sts=4:
