;
;   This is a sample IDL program that shows how an application can write a
;   congruent special value dataset. This program will write one dataset,
;   "SnowDepth", that contains actual data.
;
;   Then, this program will create another dataset, "SpecialValues". This
;   dataset only contains data for special values. This is called a "congruent"
;   dataset in the sense that this dataset will have the same rank and dimension
;   as the "SnowDepth" dataset. Note that the datatype does not need to be the
;   same as the "SnowDepth" dataset. In fact, the datatype of "SnowDepth" is
;   float while the datatype of "SpecialValues" is byte.
;
;   For simplicity, we will use a small 5-by-7 array, and "SnowDepth" will be
;   filled by meaningless numbers. We will use two special values. Although it
;   is not recorded in the file, we assume that the first special value is for
;   "WaterFilled" and its value is 1, and the second special value is for
;   "IceCovered" and its value is 2. We will not assume that two special values
;   are mutually exclusive; in other words, both special values can be set for
;   an element. In this example, an optional attribute that describes the meaning of special
;   value is added.
;
;   For more information, please refer to the following sections from RFC:
;   Special Values in HDF5 http://www.hdfgroup.org/pubs/rfcs/RFC_Special_Values_in_HDF5.pdf
;      Section 4.3.3 "Parallel Special Value Dataset"
;      Section 4.2.2 "Non-Mutually Exclusive Special Values"
;
filename = 'congruent.idl.h5' ; File name
dsetname = 'SnowDepth' ; Dataset name
svalname = 'SpecialValues' ; Parallel dataset name
attrname = 'Description' ; Attribute name
attrdesc = '1: water-filled, 2: ice-covered' ; Description of the congruent dataset

dimsize = [5, 7]

; create an HDF5 file
fileid = H5F_CREATE(filename)

; create a dataspace for a dataset and a congruent dataset
spaceid = H5S_CREATE_SIMPLE(dimsize)

; define a dataset
; fill actual data
;           0  1  2  3  4  5  6
;        +----------------------
;      0 |  1  2  3  4  5  6  7
;      1 |  8  9 10 11 12 13 14
;      2 | 15 16 17 18 19 20 21
;      3 | 22 23 24 25 26 27 28
;      4 | 29 30 31 32 33 34 35
data = FLTARR(dimsize[0], dimsize[1])
k = 1
FOR i=0,dimsize[0]-1 DO BEGIN
  FOR j=0,dimsize[1]-1 DO BEGIN
    data[i,j] = k
    k = k + 1
  ENDFOR
ENDFOR

; create a datatype
typeid = H5T_IDL_CREATE(data)

; create an HDF5 dataset that represents actual data
dsetid = H5D_CREATE(fileid, dsetname, typeid, spaceid)

; write actual data to the dataset we just created
H5D_WRITE, dsetid, data

; close the datatype
H5T_CLOSE, typeid

; close the dataset
H5D_CLOSE, dsetid

; define a congruent dataset
; fill non-mutual special values (0: no special values, 1: WaterFilled, 2: IceCovered, 3: both)
;           0  1  2  3  4  5  6
;        +----------------------
;      0 |  3  0  0  1  0  2  1
;      1 |  3  0  0  1  0  2  1
;      2 |  3  0  0  1  0  2  1
;      3 |  3  0  0  1  0  2  1
;      4 |  3  0  0  1  0  2  1
svaldata = BYTARR(dimsize[0], dimsize[1])
FOR i=0,dimsize[0]-1 DO BEGIN
  FOR j=0,dimsize[1]-1 DO BEGIN
    masks = 0
    ; for each row, special value '1' is marked once every three elements
    IF ( j MOD 3 EQ 0 ) THEN BEGIN
      masks = (masks OR 1)
    ENDIF
    ; for each row, special value '2' is marked once every five elements
    IF ( j MOD 5 EQ 0 ) THEN BEGIN
      masks = (masks OR 2)
    ENDIF
    svaldata[i,j] = masks
  ENDFOR
ENDFOR

; create a datatype
typeid = H5T_IDL_CREATE(svaldata)

; create an HDF5 dataset that represents special values for the dataset created above
dsetid = H5D_CREATE(fileid, svalname, typeid, spaceid)

; write special values to the congruent dataset
H5D_WRITE, dsetid, svaldata

; close the datatype
H5T_CLOSE, typeid

; (optional) write an attribute that describes the meaning of the special value
attrspace = H5S_CREATE_SCALAR()
attrtype = H5T_IDL_CREATE(attrdesc)
attrid = H5A_CREATE(dsetid, attrname, attrtype, attrspace)
H5A_WRITE, attrid, attrdesc
H5S_CLOSE, attrspace
H5A_CLOSE, attrid

; close the dataset
H5D_CLOSE, dsetid

; close the dataspace
H5S_CLOSE, spaceid

; close the file
H5F_CLOSE, fileid

END

; vi:set ts=4 sw=4 sts=4:
