/*
   This is a sample C program that shows how an application can read a
   congruent special value dataset. This program will read one dataset,
   "SnowDepth", that contains actual data. Value of some dataset elements
   should be replaced by special value.

   This sample code is similar to congruent_read.c. 
   To check if the n-bit filter actually reduces the file size, one can change 
   the size of dimension from {10,7} to {1000,7}.
   Since reading data using the n-bit filter is internally done by HDF5, this
   program does not need to handle any n-bit operations.
 */

#define H5_USE_16_API
#include "hdf5.h"
#include <stdio.h>
#include <stdlib.h>

#define errorexit()  _errorexit(__LINE__)

static void _errorexit(int line)
{
	fprintf(stderr, "errorexit at %d\n", line);
	exit(1);
}

static void read_file(const char *fname)
{
	hid_t fileid, spaceid, dsetid;
	hsize_t dimsize[2] = { 10, 7 };
	float *data;
	unsigned char *masks;
	int i, j;

	if ((data = malloc(sizeof(data[0]) * dimsize[0] * dimsize[1])) == NULL) errorexit();
	if ((masks = malloc(sizeof(masks[0]) * dimsize[0] * dimsize[1])) == NULL) errorexit();

	/* open the HDF5 file using read-only mode */
	if ((fileid = H5Fopen(fname, H5F_ACC_RDONLY, H5P_DEFAULT)) < 0) errorexit();

	/* create a dataspace for both the actual dataset and the congruent dataset */
	if ((spaceid = H5Screate_simple(2, dimsize, NULL)) < 0) errorexit();

	/* read a dataset */
	{
		/* open the HDF5 dataset */
		if ((dsetid = H5Dopen(fileid, "SnowDepth")) < 0) errorexit();
		/* read data from the HDF5 dataset, and store them in a variable called "data" */
		if (H5Dread(dsetid, H5T_NATIVE_FLOAT, spaceid, spaceid, H5P_DEFAULT, data) < 0) errorexit();
		/* close the dataset */
		if (H5Dclose(dsetid) < 0) errorexit();
	}

	/* read a congruent dataset */
	{
		/* open the congruent dataset */
		if ((dsetid = H5Dopen(fileid, "SpecialValues")) < 0) errorexit();
		/* read special values and store them in a variable called "masks" */
		if (H5Dread(dsetid, H5T_NATIVE_UCHAR, spaceid, spaceid, H5P_DEFAULT, masks) < 0) errorexit();
		/* close the dataset */
		if (H5Dclose(dsetid) < 0) errorexit();
	}

	/* close the dataspace */
	if (H5Sclose(spaceid) < 0) errorexit();
	/* close the file */
	if (H5Fclose(fileid) < 0) errorexit();

	/* combine the actual dataset and the congruent dataset to get a final result */
	/* this process could be done when the dataset is written, too */
	/* for each row */
	for (i = 0; i < dimsize[0]; ++i) {
		/* for each column */
		for (j = 0; j < dimsize[1]; ++j) {
			/* for each element */
			unsigned char mask;

			/* "mask" is the special value that corresponds to this data */
			mask = masks[i * dimsize[1] + j];

			/* if special value is not set for this element, just show the actual element */
			if (mask == 0)
				printf(" %4.1f ", data[i *dimsize[1] + j]);
			/* if special value is marked for this element, show a special character */
			else {
				char ch;
				/* if both special value '2' and special value '1' are marked, show 'X' */
				if (mask & 1 && mask & 2) ch = 'X';
				/* if only special value '2' is marked, show '\' */
				else if (mask & 2) ch = '\\';
				/* if only special value '1' is marked, show '/' */
				else if (mask & 1) ch = '/';
				/* if you want to merge the special value into the actual data, "data",
				 * you may want to do something like, data[i][j] = -1; */
				printf("  %c   ", ch);
			}
		}
		printf("\n");
	}

	free(data);
	free(masks);
}

int main(int argc, char **argv)
{
	read_file("nbit.c.h5");
	return 0;
}

/* vi:set ts=4 sw=4 sts=4: */
