/*
   This is a sample C program that shows how an application can read attribute
   triplets that include special value information.  This program will read one
   dataset, "SnowDepth", that contains actual data. Value of some dataset
   elements should be special value for the real world user.  In this example,
   this dataset has an attribute, "SpecialValues", which contains an array of
   triplets that describe the special value information.

   This program will read the "SpecialValue" attribute. We assume that
   there are two special values; 1 represents "WaterFilled" and 2 represents
   "IceCovered".

   After reading the dataset and the attribute, this program will print some
   human-readable output. First, this program will print regions where a
   special value applies to. Then, it will traverse each element, and print
   either value or special mark. If none of special values are set for an
   element, the actual value stored in the "SnowDepth" will be printed. If some
   special values are set, a special character will be printed instead. In this
   example, '/' is for the "WaterFilled" special value, '\' is for the
   "IceCovered" special value, and 'X' is for the overlapped special value.

   For more information, please refer to the following sections from RFC:
   Special Values in HDF5 http://www.hdfgroup.org/pubs/rfcs/RFC_Special_Values_in_HDF5.pdf
      Section 4.3.4 "Attribute Triplet"
      Section 4.2.2 "Non-Mutually Exclusive Special Values"
 */
#define H5_USE_16_API
#include "hdf5.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define DATASET_RANK  2
#define NUM_SPECIALVALUES  2

typedef struct {
	const char *description;
	int value;
	/* the following type corresponds to H5R_DATASET_REGION */
	hdset_reg_ref_t region;
} triplet_t;

#define errorexit()  _errorexit(__LINE__)

static void _errorexit(int line)
{
	fprintf(stderr, "errorexit at %d\n", line);
	exit(1);
}

static void read_file(const char *fname)
{
	hid_t fileid, spaceid, dsetid;
	hsize_t dimsize[DATASET_RANK] = { 5, 7 };
	float data[5][7];
	int masks[5][7];
	int i, j;

	/* open the HDF5 file using read-only mode */
	if ((fileid = H5Fopen(fname, H5F_ACC_RDONLY, H5P_DEFAULT)) < 0) errorexit();

	/* create a dataspace for the dataset */
	if ((spaceid = H5Screate_simple(DATASET_RANK, dimsize, NULL)) < 0) errorexit();

	/* open the HDF5 dataset */
	if ((dsetid = H5Dopen(fileid, "SnowDepth")) < 0) errorexit();

	/* read data from the HDF5 dataset, and store them in a variable called "data" */
	if (H5Dread(dsetid, H5T_NATIVE_FLOAT, spaceid, spaceid, H5P_DEFAULT, data) < 0) errorexit();

	memset(masks, 0, sizeof(masks));
	/* read special values */
	{
		hid_t desc_tid, triplet_tid, sv_aid;
		triplet_t buf[NUM_SPECIALVALUES];

		/* create a compound datatype for a special value */
		if ((triplet_tid = H5Tcreate(H5T_COMPOUND, sizeof(triplet_t))) < 0) errorexit();
		if ((desc_tid = H5Tcopy(H5T_C_S1)) < 0) errorexit();
		if (H5Tset_size(desc_tid, H5T_VARIABLE) < 0) errorexit();
		if (H5Tinsert(triplet_tid, "description", HOFFSET(triplet_t, description), desc_tid) < 0) errorexit();
		if (H5Tinsert(triplet_tid, "value", HOFFSET(triplet_t, value), H5T_NATIVE_INT) < 0) errorexit();
		if (H5Tinsert(triplet_tid, "region", HOFFSET(triplet_t, region), H5T_STD_REF_DSETREG) < 0) errorexit();

		/* open the attribute for special values */
		if ((sv_aid = H5Aopen(dsetid, "SpecialValues", H5P_DEFAULT)) < 0) errorexit();

		/* read special values */
		if (H5Aread(sv_aid, triplet_tid, buf) < 0) errorexit();

		/* iterate all special values */
		for (i = 0; i < NUM_SPECIALVALUES; ++i) {
			hid_t maskedspace;
			const triplet_t *triplet = &buf[i];
			hssize_t numblock;

			/* get the dataspace corresponding to the region that uses special value */
			if ((maskedspace = H5Rget_region(dsetid, H5R_DATASET_REGION, &triplet->region)) < 0) errorexit();

			/* get the number of hyperslab blocks in the region */
			if ((numblock = H5Sget_select_hyper_nblocks(maskedspace)) < 0) errorexit();
			for (j = 0; j < (int)numblock; ++j) {
				/* assumes that the rank is 2 (DATASET_RANK). the rank can be retrieved by H5Sget_simple_extent_dims */
				hsize_t corners[2][DATASET_RANK], k, l;
				/* read two corners of one hyperslab block defined by the region */
				if (H5Sget_select_hyper_blocklist(maskedspace, j, 1, (hsize_t *)corners) < 0) errorexit();
				printf(" %s's region %d : (%lu, %lu) ~ (%lu, %lu)\n", triplet->description, j + 1, (unsigned long)corners[0][0], (unsigned long)corners[0][1], (unsigned long)corners[1][0], (unsigned long)corners[1][1]);

				/* mark all elements that belong to this hyperslab block
				 *  (corners[0][0], corners[0][1]) ......................
				 *  .....................................................
				 *  ....................... (corners[1][0][corners[1][1])
				 */
				for (k = corners[0][0]; k <= corners[1][0]; ++k) {
					for (l = corners[0][1]; l <= corners[1][1]; ++l)
						masks[k][l] |= triplet->value;
				}
			}

			if (H5Sclose(maskedspace) < 0) errorexit();
		}

		/* close the attribute */
		if (H5Aclose(sv_aid) < 0) errorexit();
		/* close the datatype for a special value */
		if (H5Tclose(triplet_tid) < 0) errorexit();
		if (H5Tclose(desc_tid) < 0) errorexit();
	}

	/* close the dataset */
	if (H5Dclose(dsetid) < 0) errorexit();
	/* close the dataspace */
	if (H5Sclose(spaceid) < 0) errorexit();
	/* close the file */
	if (H5Fclose(fileid) < 0) errorexit();

	/* combine the actual dataset with the special value stored in triplet to get a final result */
	/* this process could be done when the dataset is written, too */
	/* for each row */
	for (i = 0; i < 5; ++i) {
		/* for each column */
		for (j = 0; j < 7; ++j) {
			/* for each element */
			int mask;

			/* "mask" is the special value that corresponds to this data */
			mask = masks[i][j];

			/* if special value is not set for this element, just show the actual element */
			if (mask == 0)
				printf(" %4.1f ", data[i][j]);
			/* if special value is marked for this element, show a special character */
			else {
				char ch;
				/* if both special value '2' and special value '1' are marked, show 'X' */
				if (mask & 1 && mask & 2) ch = 'X';
				/* if only special value '2' is marked, show '\' */
				else if (mask & 2) ch = '\\';
				/* if only special value '1' is marked, show '/' */
				else if (mask & 1) ch = '/';
				/* if you want to merge the special value into the actual data, "data",
				 * you may want to do something like, data[i][j] = -1; */
				printf("  %c   ", ch);
			}
		}
		printf("\n");
	}
}

int main(int argc, char **argv)
{
	read_file("regionsref.c.h5");
	return 0;
}

/* vi:set ts=4 sw=4 sts=4: */
